@@ -1,13 +1,15 @@
+from typing import Literal
+
+import matplotlib.colors as mcolors
 import matplotlib.pyplot as plt
 import pandas as pd
 import plotly.graph_objects as go
 from plotly.graph_objs.graph_objs import FigureWidget
 from plotly.subplots import make_subplots
 
-from qubx.utils.charting.mpl_helpers import ohlc_plot
-from qubx.utils.charting.mpl_helpers import plot_trends
-from qubx.utils.charting.mpl_helpers import subplot
+from qubx.core.series import OHLCV, TimeSeries
 from qubx.utils import Struct
+from qubx.utils.charting.mpl_helpers import ohlc_plot, plot_trends, subplot
 
 
 def install_plotly_helpers():
@@ -91,7 +93,21 @@ def install_plotly_helpers():
                 )
             )
 
-        def arrow(look, x2, y2, x1, y1, c="red", text="", lw=1, font=dict(size=8)):
+        def hline(look, y, c="yellow", lw=1, ls="dot"):
+            return look.add_shape(
+                go.layout.Shape(
+                    type="line",
+                    x0=0,
+                    x1=1,
+                    y0=y,
+                    y1=y,
+                    xref="paper",
+                    yref="y1",
+                    line=dict(width=lw, color=c, dash=ls),
+                )
+            )
+
+        def arrow(look, x2, y2, x1, y1, c="red", text="", lw=1, font=dict(size=8), head=1):
             return look.add_annotation(
                 x=x1,
                 y=y1,
@@ -104,25 +120,28 @@ def install_plotly_helpers():
                 text=text,
                 font=font,
                 showarrow=True,
-                arrowhead=1,
+                arrowhead=head,
                 arrowsize=1,
                 arrowwidth=lw,
                 arrowcolor=c,
             )
 
-        def custom_hover(v, h=600, n=2, legend=False, show_info=True):
+        def hover(v, h=600, n=2, legend=False, show_info=True, w=None):
             return (
                 v.update_traces(xaxis="x1")
                 .update_layout(
                     height=h,
+                    width=w,
                     hovermode="x unified",
                     showlegend=legend,
-                    hoverdistance=1000 if show_info else 0,
+                    hoverdistance=1 if show_info else 0,
                     xaxis={"hoverformat": "%d-%b-%y %H:%M"},
                     yaxis={"hoverformat": f".{n}f"},
                     dragmode="zoom",
                     newshape=dict(line_color="yellow", line_width=1.0),
                     modebar_add=["drawline", "drawopenpath", "drawrect", "eraseshape"],
+                    # hoversubplots="axis",
+                    hoverlabel=dict(align="auto", bgcolor="rgba(10, 10, 10, 0.5)"),
                 )
                 .update_xaxes(
                     showspikes=True,
@@ -140,14 +159,15 @@ def install_plotly_helpers():
                 )
             )
 
-        FigureWidget.hover = custom_hover  # type: ignore
+        FigureWidget.hover = hover  # type: ignore
         FigureWidget.rline = rline  # type: ignore
         FigureWidget.rlinex = rlinex  # type: ignore
         FigureWidget.rline_ = rline_  # type: ignore
         FigureWidget.vline = vline  # type: ignore
+        FigureWidget.hline = hline  # type: ignore
         FigureWidget.dliney = dliney  # type: ignore
         FigureWidget.arrow = arrow  # type: ignore
-    except:
+    except:  # noqa: E722
         print(" >>> Cant attach helpers to plotly::FigureWidget - probably it isn't installed !")
 
 
@@ -243,8 +263,14 @@ class AbstractLookingGlass:
 
 
 class LookingGlass:
-
-    def __init__(self, master, studies: dict | None = None, title="", backend="plotly", **kwargs):
+    def __init__(
+        self,
+        master,
+        studies: dict | None = None,
+        title="",
+        backend: Literal["matplotlib", "mpl", "plotly", "ply", "plt"] = "plotly",
+        **kwargs,
+    ):
         if backend in ["matplotlib", "mpl"]:
             self.__instance = LookingGlassMatplotLib(master=master, studies=studies, title=title, **kwargs)
         elif backend in ["plotly", "ply", "plt"]:
@@ -257,7 +283,6 @@ class LookingGlass:
 
 
 class LookingGlassMatplotLib(AbstractLookingGlass):
-
     def __init__(
         self,
         master,
@@ -276,8 +301,10 @@ class LookingGlassMatplotLib(AbstractLookingGlass):
         self._fmt = fmt
         self._ohlc_width = ohlc_width
         self._n_style = "-"
+        self._last_color = None
+        self._last_mark = None
 
-    def __plt_series(self, y, zoom, study_name, k, plot_style="line"):
+    def _plt_series(self, y, zoom, study_name, k, plot_style="line"):
         _forced_limits = []
         if isinstance(y, (int, float)):
             try:
@@ -287,7 +314,8 @@ class LookingGlassMatplotLib(AbstractLookingGlass):
         else:
             _lbl = y.name if hasattr(y, "name") and y.name else ("%s_%d" % (study_name, k))
 
-            if isinstance(y, pd.DataFrame):
+            if isinstance(y, (pd.DataFrame, OHLCV)):
+                y = y.pd() if isinstance(y, OHLCV) else y
 
                 yy = y[zoom] if zoom else y
 
@@ -427,29 +455,65 @@ class LookingGlassMatplotLib(AbstractLookingGlass):
                     plt.plot(yy["close"], lw=0, label=_lbl)
                 else:
                     for _col in yy.columns:
-                        self.__plot_as_type(yy[_col], plot_style, self._n_style, _col)
+                        self._plot_as_type(yy[_col], plot_style, self._n_style, _col)
             else:
+                y = y.pd() if isinstance(y, TimeSeries) else y
                 yy = y[zoom] if zoom else y
-                self.__plot_as_type(yy, plot_style, self._n_style, _lbl)
+                self._plot_as_type(yy, plot_style, self._n_style, _lbl)
 
             # we want to see OHLC at maximal scale
             return _forced_limits
 
-    def __plot_as_type(self, y, plot_style, line_style: str, label):
-        __clr = line_style[0] if len(line_style) > 0 and line_style[0].isalpha() else None
+    def _extract_colors_marks(self, line: str):
+        from matplotlib.markers import MarkerStyle
+
+        mplc = mcolors.get_named_colors_mapping()
+
+        colors, marks = [], []
+        r_l = ""
+        for c in (line if line else "").split(" "):
+            if c.startswith("#") or c.lower() in mplc:
+                colors.append(c)
+                continue
+            if c.startswith("arrow-") or c in MarkerStyle.markers:
+                marks.append("^" if c.endswith("arrow-up") else "v" if c.endswith("arrow-down") else c)
+                continue
+            if c.startswith("dots") or c in MarkerStyle.markers:
+                marks.append(".")
+                continue
+            r_l += c
+        return r_l if r_l else None, colors, marks
+
+    def _plot_as_type(self, y, plot_style, line_style: str, label):
+        _l_style, clrs, mrks = self._extract_colors_marks(line_style)
+        _clr = clrs[0] if clrs else None
+        if _clr:
+            self._last_color = _clr
+
         if plot_style == "line":
-            plt.plot(y, line_style, label=label)
+            _l_style = "" if self._last_mark else _l_style
+            plt.plot(
+                y,
+                ls=_l_style,
+                label=label,
+                color=self._last_color,
+                marker=self._last_mark,
+                ms=1 if self._last_mark == "." else 6,
+            )
+
         elif plot_style == "area":
-            plt.fill_between(y.index, y, color=__clr, label=label)
+            plt.fill_between(y.index, y, color=_clr, label=label)
         elif plot_style.startswith("step"):
             _where = "post" if "post" in plot_style else "pre"
-            plt.step(y.index, y, color=__clr, where=_where, label=label)
+            plt.step(y.index, y, color=_clr, where=_where, label=label)
         elif plot_style.startswith("bar"):
-            _bw = pd.Series(y.index).diff().mean().total_seconds() / 24 / 60 / 60
-            plt.bar(y.index, y, lw=0.4, width=_bw, edgecolor=__clr, color=__clr, label=label)
+            _bw = pd.Series(y.index).diff().mean().total_seconds() / 24 / 60 / 60  # type: ignore
+            plt.bar(y.index, y, lw=0.4, width=_bw, edgecolor=_clr, color=_clr, label=label)
+
+        # - reset last mark
+        self._last_mark = None
 
     def _show_plot(self, vert_bar, title, zoom):
-
         # plot all master series
         shape = (self.s_size * len(self.s) + self.m_size, 1)
         subplot(shape, 1, rowspan=self.m_size)
@@ -459,9 +523,15 @@ class LookingGlassMatplotLib(AbstractLookingGlass):
         for j, m in enumerate(ms):
             # if style description
             if isinstance(m, str):
-                self._n_style = m
+                _l_style, clrs, mrks = self._extract_colors_marks(m)
+                _clr = clrs[0] if clrs else None
+                _marker = mrks[0] if mrks else None
+                self._n_style = _l_style
+                self._last_color = _clr if _clr else self._last_color
+                self._last_mark = _marker if _marker else self._last_mark
+
             else:
-                _lims = self.__plt_series(m, zoom, "Master", j)
+                _lims = self._plt_series(m, zoom, "Master", j)
                 self._n_style = "-"
                 if _limits_to_set is None and _lims:
                     _limits_to_set = _lims
@@ -491,7 +561,6 @@ class LookingGlassMatplotLib(AbstractLookingGlass):
 
             wait_for_limits = False
             for j, v in enumerate(vs):
-
                 # if we need to read limits
                 if wait_for_limits:
                     if isinstance(v, (list, tuple)) and len(v) > 1:
@@ -509,7 +578,7 @@ class LookingGlassMatplotLib(AbstractLookingGlass):
                     else:
                         self._n_style = v
                 else:
-                    self.__plt_series(v, zoom, k, j, plot_style=plot_style)
+                    self._plt_series(v, zoom, k, j, plot_style=plot_style)
 
             if vert_bar:
                 plt.axvline(vert_bar, ls="-.", lw=0.5)
@@ -519,6 +588,39 @@ class LookingGlassMatplotLib(AbstractLookingGlass):
 
         self._n_style = "-"
 
+        class EmptyPlot:
+            def hover(self, *args, **kwargs):
+                dpi = plt.rcParams["figure.dpi"]
+
+                h = kwargs.get("h") or kwargs.get("height")
+                w = kwargs.get("w") or kwargs.get("width")
+
+                if h:
+                    h = max(1, h / dpi)
+                if w:
+                    w = max(1, w / dpi)
+
+                if h and not w:
+                    w = 1.618 * h
+                elif w and not h:
+                    h = 0.618 * w
+
+                if h and w:
+                    plt.gcf().set_size_inches(w, h)
+
+                return self
+
+            def show(self, *args, **kwargs):
+                pass
+
+            def __str__(self):
+                return title if title else ""
+
+            def __repr__(self):
+                return title if title else ""
+
+        return EmptyPlot()
+
 
 class LookingGlassPlotly(AbstractLookingGlass):
     TREND_COLORS = Struct(
@@ -544,7 +646,25 @@ class LookingGlassPlotly(AbstractLookingGlass):
     def __plt_series(self, y, zoom, study_name, k, row, col, plot_style="line"):
         _lbl = y.name if hasattr(y, "name") and y.name else ("%s_%d" % (study_name, k))
 
-        if isinstance(y, pd.DataFrame):
+        def _scatter(
+            xs: pd.Series, comments: pd.Series | None, name: str, marker: str, color: str, size: int = 11
+        ) -> None:
+            _args = dict(
+                x=xs.index,
+                y=xs,
+                mode="markers",
+                name=name,
+                text=comments,
+                marker={
+                    "symbol": marker,
+                    "size": size,
+                    "color": color,
+                },
+            )
+            self.fig.add_trace(go.Scatter(**_args), row=row, col=col)
+
+        if isinstance(y, (pd.DataFrame, OHLCV)):
+            y = y.pd() if isinstance(y, OHLCV) else y
             yy = y[zoom] if zoom else y
 
             # candlesticks
@@ -760,70 +880,47 @@ class LookingGlassPlotly(AbstractLookingGlass):
             elif self._frame_has_cols(y, ["Type", "Time", "Price", "PriceOccured"]):
                 _bot = yy[yy.Type == "-"]
                 _top = yy[yy.Type == "+"]
-                self.fig.add_trace(
-                    go.Scatter(
-                        x=_bot.index,
-                        y=_bot.PriceOccured,
-                        mode="markers",
-                        name="B",
-                        marker={"symbol": "triangle-right"},
-                    ),
-                    row=row,
-                    col=col,
-                )
 
-                self.fig.add_trace(
-                    go.Scatter(
-                        x=_top.index,
-                        y=_top.PriceOccured,
-                        mode="markers",
-                        name="T",
-                        marker={"symbol": "triangle-right"},
-                    ),
-                    row=row,
-                    col=col,
-                )
+                _scatter(_bot.PriceOccured, None, "B", "triangle-right", "#3cfa00", 12)
+                _scatter(_top.PriceOccured, None, "T", "triangle-right", "#3cfa00", 12)
 
-            # executed signals from signal tester
+            # executions from executor's log
             elif self._frame_has_cols(y, ["exec_price", "quantity"]):
                 _b_ords = yy[yy.quantity > 0]
                 _s_ords = yy[yy.quantity < 0]
-                self.fig.add_trace(
-                    go.Scatter(
-                        x=_b_ords.index,
-                        y=_b_ords.exec_price,
-                        mode="markers",
-                        name="BOT",
-                        marker={
-                            "symbol": "triangle-up",
-                            "size": 13,
-                            "color": "#3cfa00",
-                        },
-                    ),
-                    row=row,
-                    col=col,
-                )
 
-                self.fig.add_trace(
-                    go.Scatter(
-                        x=_s_ords.index,
-                        y=_s_ords.exec_price,
-                        mode="markers",
-                        name="SLD",
-                        marker={
-                            "symbol": "triangle-down",
-                            "size": 13,
-                            "color": "#20ffff",
-                        },
-                    ),
-                    row=row,
-                    col=col,
-                )
+                # - how much was traded
+                _b_info = "<i>bought</i> " + _b_ords.quantity.astype(str)
+                _s_info = "<i>sold</i> " + _s_ords.quantity.astype(str)
+
+                _scatter(_b_ords.exec_price, _b_info, "BOT", "triangle-up", "#25ff00", 12)
+                _scatter(_s_ords.exec_price, _s_info, "SLD", "triangle-down", "#f0d", 12)
+                # _scatter(_s_ords.exec_price, _s_info, "SLD", "triangle-down", "#20ffff", 12)
+
+            # 27-aug-2024: show generated signals from Qubx backtester
+            elif self._frame_has_cols(y, ["signal", "reference_price", "price", "take", "stop", "group", "comment"]):
+                _sel_price = lambda x: x["reference_price"].where(x["price"].isna(), x["price"])  # noqa: E731
+                _b_sigs = yy[yy.signal > 0]
+                _z_sigs = yy[yy.signal == 0]
+                _s_sigs = yy[yy.signal < 0]
+                # - Longs
+                _scatter(_sel_price(_b_sigs), _b_sigs["comment"], "LONG", "triangle-up-open-dot", "#3cfa00")
+                _scatter(_b_sigs[~_b_sigs["take"].isna()]["take"], None, "Take", "line-ew-open", "#3cfa00")
+                _scatter(_b_sigs[~_b_sigs["stop"].isna()]["stop"], None, "Stop", "line-ew-open", "#fa3c00")
+
+                # - Shorts
+                _scatter(_sel_price(_s_sigs), _s_sigs["comment"], "SHORTS", "triangle-down-open-dot", "#20ffff")
+                _scatter(_s_sigs[~_s_sigs["take"].isna()]["take"], None, "Take", "line-ew-open", "#3cfa00")
+                _scatter(_s_sigs[~_s_sigs["stop"].isna()]["stop"], None, "Stop", "line-ew-open", "#fa3c00")
+
+                # - Exits
+                _scatter(_sel_price(_z_sigs), _z_sigs["comment"], "EXITS", "circle-open-dot", "#ffffff")
 
             else:
                 for _col in yy.columns:
                     self.__plot_as_type(yy[_col], row, col, plot_style, _col)
         else:
+            y = y.pd() if isinstance(y, TimeSeries) else y
             yy = y[zoom] if zoom else y
             self.__plot_as_type(yy, row, col, plot_style, _lbl)
 
@@ -832,37 +929,20 @@ class LookingGlassPlotly(AbstractLookingGlass):
         if plot_style == "line":
             self.fig.add_trace(
                 go.Scatter(
-                    x=y.index,
-                    y=y,
-                    mode="lines",
-                    line={"width": 0.5, "dash": style, "color": color},
-                    name=label,
+                    x=y.index, y=y, mode="lines", line={"width": 0.5, "dash": style, "color": color}, name=label
                 ),
                 row=row,
                 col=col,
             )
         elif plot_style == "area":
             self.fig.add_trace(
-                go.Scatter(
-                    x=y.index,
-                    y=y,
-                    mode="lines",
-                    line={"width": 1, "color": color},
-                    fill="tozeroy",
-                    name=label,
-                ),
+                go.Scatter(x=y.index, y=y, mode="lines", line={"width": 1, "color": color}, fill="tozeroy", name=label),
                 row=row,
                 col=col,
             )
         elif plot_style.startswith("step"):
             self.fig.add_trace(
-                go.Scatter(
-                    x=y.index,
-                    y=y,
-                    mode="lines",
-                    line={"shape": "hv", "color": color},
-                    name=label,
-                ),
+                go.Scatter(x=y.index, y=y, mode="lines", line={"shape": "hv", "color": color}, name=label),
                 row=row,
                 col=col,
             )
@@ -910,8 +990,7 @@ class LookingGlassPlotly(AbstractLookingGlass):
 
         plotly_style, plotly_color = None, None
 
-        if mpl_style in plotly_styles:
-            # specified plotly style line
+        if mpl_style in plotly_styles:  # specified plotly style line
             plotly_style = mpl_style
         elif mpl_style == ":":
             plotly_style = "dot"
@@ -922,8 +1001,7 @@ class LookingGlassPlotly(AbstractLookingGlass):
         elif mpl_style == "-":
             plotly_style = "solid"
 
-        if plotly_style is None and mpl_color is None:
-            # it looks only color is specified
+        if plotly_style is None and mpl_color is None:  # it looks only color is specified
             mpl_color = mpl_style
 
         if mpl_color == "r":
@@ -937,14 +1015,14 @@ class LookingGlassPlotly(AbstractLookingGlass):
 
         return plotly_style, plotly_color
 
-    def __mpl_color_to_plotly(self, color):
-        if color == "r":
-            color = "red"
-        elif color == "g":
-            color = "green"
-        elif color == "w":
-            color = "white"
-        return color
+    # def __mpl_color_to_plotly(self, color):
+    #     if color == "r":
+    #         color = "red"
+    #     elif color == "g":
+    #         color = "green"
+    #     elif color == "w":
+    #         color = "white"
+    #     return color
 
     def _show_plot(self, vert_bar, title, zoom):
         # plot all master series
@@ -953,12 +1031,9 @@ class LookingGlassPlotly(AbstractLookingGlass):
         axis_rules = {}
         if len(self.s):
             row_heights.extend([(1 - master_fraction) / len(self.s)] * len(self.s))
+
         self.fig = make_subplots(
-            rows=len(self.s) + 1,
-            cols=1,
-            shared_xaxes=True,
-            vertical_spacing=0.01,
-            row_heights=row_heights,
+            rows=len(self.s) + 1, cols=1, shared_xaxes=True, vertical_spacing=0.01, row_heights=row_heights
         )
 
         ms = self.m if isinstance(self.m, (tuple, list)) else [self.m]
@@ -1031,27 +1106,16 @@ class LookingGlassPlotly(AbstractLookingGlass):
                 else:
                     self.__plt_series(v, zoom, k, j, i, 1, plot_style=plot_style)
 
-        if title is not None:
-            plot_title = "%s %s" % (self._title, str(title))
-        else:
-            plot_title = "%s %s" % (self._title, str(vert_bar))
-
-        if plot_title is not None:
+        plot_title = title or self._title or ""
+        plot_title += str(vert_bar) if vert_bar else ""  # add title to plot
+        if plot_title:
             self.fig.update_layout(
-                title={
-                    "text": plot_title,
-                    "y": 0.9,
-                    "x": 0.5,
-                    "xanchor": "center",
-                    "yanchor": "top",
-                }
+                title={"text": plot_title, "y": 0.9, "x": 0.5, "xanchor": "center", "yanchor": "top"}
             )
 
         self.fig.update_layout(axis_rules)
         self.fig.update_layout(
-            xaxis_rangeslider_visible=False,
-            margin=dict(l=5, r=5, t=35, b=5),
-            height=self.mph + self.sph * len(self.s),
+            xaxis_rangeslider_visible=False, margin=dict(l=5, r=5, t=35, b=5), height=self.mph + self.sph * len(self.s)
         )
 
         return go.FigureWidget(self.fig)
@@ -1060,14 +1124,7 @@ class LookingGlassPlotly(AbstractLookingGlass):
         # Line Vertical
         self.fig.add_shape(
             go.layout.Shape(
-                type="line",
-                x0=x,
-                x1=x,
-                xref=xref,
-                yref="paper",
-                y0=0,
-                y1=1,
-                line=dict(width=1, dash="dot"),
+                type="line", x0=x, x1=x, xref=xref, yref="paper", y0=0, y1=1, line=dict(width=1, dash="dot")
             )
         )
 
